import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;


import org.waveprotocol.wave.examples.fedone.common.DocumentConstants;
import org.waveprotocol.wave.examples.fedone.common.HashedVersion;
import org.waveprotocol.wave.examples.fedone.util.BlockingSuccessFailCallback;
import org.waveprotocol.wave.examples.fedone.waveclient.common.ClientBackend;
import org.waveprotocol.wave.examples.fedone.waveclient.common.ClientUtils;
import org.waveprotocol.wave.examples.fedone.waveclient.common.ClientWaveView;
import org.waveprotocol.wave.examples.fedone.waveclient.common.IndexEntry;
import org.waveprotocol.wave.examples.fedone.waveclient.common.WaveletOperationListener;
import org.waveprotocol.wave.examples.fedone.waveserver.WaveClientRpc.ProtocolSubmitResponse;
import org.waveprotocol.wave.model.document.operation.AnnotationBoundaryMap;
import org.waveprotocol.wave.model.document.operation.Attributes;
import org.waveprotocol.wave.model.document.operation.BufferedDocOp;
import org.waveprotocol.wave.model.document.operation.DocInitializationCursor;
import org.waveprotocol.wave.model.document.operation.impl.InitializationCursorAdapter;
import org.waveprotocol.wave.model.operation.wave.AddParticipant;
import org.waveprotocol.wave.model.operation.wave.RemoveParticipant;
import org.waveprotocol.wave.model.operation.wave.WaveletDocumentOperation;
import org.waveprotocol.wave.model.wave.ParticipantId;
import org.waveprotocol.wave.model.wave.data.WaveletData;

/**
 * Diese Klasse ist zentrale Bestandteil und regelt im Wesentlichen die Kommunikation zwischen Clientbackend und den GUIs.
 * 
  * @author Jan Schwarzer
 *
 */
public class Client implements WaveletOperationListener {
	
	// Standard Konstruktor
	public Client(){}
	
	// Schalten für die Ausgabe von Debuginformationen
	public boolean debug = true;

	// Hilfsvariable für die Debugausgaben
	private int debugCounter = 1;

	// Das eigentliche Backend von Google Wave
	public ClientBackend backend = null;

	// Aktuell geöffnete Wave
	public ClientWaveView openWave = null;
	
	// Liste meiner vorhandenen Waves
	List<IndexEntry> myWaves = null;
	
	// Instanz meiner Client-GUI
	public ClientGUI clientGUI = null;
	
	// Meine Runnable-Komponente um die GUI zu aktualisieren
	public MyRunnable myRunnable = null;
	
	/**
	 * Klasse für das Ausführen von Operationen auf der Client-GUI
	 * 
	 * @author Jan Schwarzer
	 *
	 */
	private class MyRunnable implements Runnable {

		private String filterOrUpdateText = null;
		
		public MyRunnable(String filterOrUpdateText){
			this.filterOrUpdateText = filterOrUpdateText;
		}

		@Override
		public void run() {
			// GUI-Initialisierung
			if(filterOrUpdateText.equals("initializeGUI")){
				// 1. Ermitteln der eigenen Kontakte (noch statisch)
				clientGUI.listContacts.add("ernstpeter");
				clientGUI.listContacts.add("fritz");
				
				// 2. Anzeigen meiner Waves
				getMyWaves();
			
			// Hinzufügen der Teilnehmer einer Wave
			} else if(filterOrUpdateText.equals("setParticipants")) {
				clientGUI.listParticipants.removeAll();
				
				for(ParticipantId participant : ClientUtils.getConversationRoot(openWave).getParticipants()) {
					clientGUI.listParticipants.add(transformAdress(participant.getAddress()));
				}
			
			// Hinzufügen der Inhalte einer Wave
			} else if(filterOrUpdateText.equals("setContent")) {
				// Liste aller BLips und Reihenfolge der Diskussion
			    Map<String, BufferedDocOp> documentMap = ClientUtils.getConversationRoot(openWave).getDocuments();
			    // Reihenfolge der Diskussion, extrahiert aus der dokumentMap
			    BufferedDocOp manifest = documentMap.get("conversation");

			    // Vor dem Auslesen der aktuellen Wave-Inhalte wird das GUI-Feld "textWaveContent" entleert
			    clientGUI.textWaveContent.setText("");

				// Für jede enthaltene Nachricht, wird das GUI-Feld "textWaveContent" aktualisiert
				for(String entry : getWaveletData(manifest, documentMap)) {
					clientGUI.textWaveContent.append(entry);
				}
			}
		}
		
		/**
		 * Diese Methode liefert zu jedem Eintrag (BlipID) in document den Blip-Klartext in Form einer String-Liste zurück.
		 * 
		 * @param document
		 * @param documentMap
		 * @return
		 */
		private List<String> getWaveletData(BufferedDocOp document, final Map<String, BufferedDocOp> documentMap) {
			final List<String> entries = new ArrayList<String>();
			
			document.apply(new InitializationCursorAdapter(
				new DocInitializationCursor() {
					
					@Override
					public void characters(String s) {}
					
					@Override
					public void annotationBoundary(AnnotationBoundaryMap map){}
					
					@Override
					public void elementEnd(){}

					@Override
					public void elementStart(String type, Attributes attrs) {
						// Nur wenn der type einer Blip entspricht
						if (type.equals(DocumentConstants.BLIP)) {
							if (attrs.containsKey(DocumentConstants.BLIP_ID)) {
								// Meiner Liste füge ich den konkreten Klartext aus documentMap hinzu
								entries.add(documentMap.get(attrs.values().iterator().next()).getCharactersString(3) + "\n");
							}
						}
					}
			}));
			
			return entries;
		}
	}
	
	/**
	 * Diese Methode sorgt für den Aufbau einer Verbindung zum Google Wave Server
	 * 
	 * @param userAtDomain
	 * @param server
	 * @param portString
	 * @throws ServerConnectException
	 */
	public void connectToBackend(String userAtDomain, String server, String portString) throws ServerConnectException {
		// Eine Verbindung nur aufbauen, wenn diese nicht bereits vorhanden ist
		if (!clientIsConnected()) {
			int port;
			
			// Nur gültige Ports zulassen
			try {
				port = Integer.parseInt(portString);
			} catch (NumberFormatException e) {
				throw new ServerConnectException("Bitte geben Sie einen gültigen Port ein!");
			}

			// Ein neues ClientBackend instanziieren
			try {
				backend = new ClientBackend(userAtDomain, server, port);
				
				if (debug)
					System.out
							.println("\n--------------------------------------  DEBUG INFO START (connectToBackend)  --------------------------------------"
									+ "\nVerbindung zu " + server + ", mit Port " + portString + " und unter dem Account " + userAtDomain + " wurde hergestellt.\n"
									+ "--------------------------------------  DEBUG INFO END (connectToBackend)  --------------------------------------\n");

				// Meinen Client am Backend registrieren
				backend.addWaveletOperationListener(this);
			} catch (Exception e) {
				throw new ServerConnectException("Verbindung konnte nicht aufgebaut werden!");
			}
		} else {
			throw new ServerConnectException("Es besteht bereits eine Verbindung zum Google Wave Server!");
		}
	}
	
	/**
	 * Diese Methode wertet die eingehenden Kommandos aus und ruft entsprechend weitere Methoden auf
	 * 
	 * @param command
	 * @param additional
	 * @param args
	 */
	public void doCommand(String command, String additional) {
		// Eine neue Wave anlegen
		if (command.equals("new")) {
			if(backend != null) {
				BlockingSuccessFailCallback<ProtocolSubmitResponse, String> callback = BlockingSuccessFailCallback.create();
				backend.createConversationWave(callback);
				callback.await(1, TimeUnit.MINUTES);
			
				if (debug)
					System.out
							.println("\n--------------------------------------  DEBUG INFO START (doCommand, new)  --------------------------------------"
									+ "\nNeue Wave erstellt.\n"
									+ "--------------------------------------  DEBUG INFO END (doCommand, new)  --------------------------------------\n");
			}
			
		// Eine Wave öffnen
		} else if (command.equals("open")) {
			String waveToOpen = clientGUI.listWaveList.getItem(clientGUI.listWaveList.getSelectionIndex());

			// Nur wenn ich eine Wave wirklich öffnen möchte, soll fortgefahren werden
			if(waveToOpen.contains("w+")) {
				int waveToOpenIndex = clientGUI.listWaveList.getSelectionIndex();
				
				if (this.openWave == null || !this.openWave.getWaveId().getId().equals(myWaves.get(waveToOpenIndex).getWaveId().getId())) {
					// Setzen der geöffneten Wave
					openWave = backend.getWave(myWaves.get(waveToOpenIndex).getWaveId());
				
					// Wenn eine Wave erfolgreich geöffnet wurde, dann selektieren wir dessen Inhalte und Teilnehmer
					if(openWave != null) {
						// Setzen der Teilnehmer einer Wave
						myRunnable.filterOrUpdateText = "setParticipants";
						myRunnable.run();
	
						// Setzen des Inhalts einer Wave
						myRunnable.filterOrUpdateText = "setContent";
						myRunnable.run();
	
						if (debug)
							System.out
									.println("\n--------------------------------------  DEBUG INFO START (doCommand, open)  --------------------------------------"
											+ "\n" + waveToOpen + " geöffnet.\n"
											+ "--------------------------------------  DEBUG INFO END (doCommand, open)  --------------------------------------\n");
					}
				} else {
					if (debug)
						System.out
								.println("\n--------------------------------------  DEBUG INFO START (doCommand, open)  --------------------------------------"
										+ "\n" + waveToOpen + " ist bereits geöffnet.\n"
										+ "--------------------------------------  DEBUG INFO END (doCommand, open)  --------------------------------------\n");
				}

			}
			
		// Eine Nachricht senden
		} else if (command.equals("send")) {
			if(this.openWave != null) {
				backend.sendAndAwaitWaveletDelta(getOpenWavelet().getWaveletName(),
						ClientUtils.createAppendBlipDelta(getManifestDocument(),
								backend.getUserId(), backend.getIdGenerator()
										.newDocumentId(), transformAdress(backend.getUserId().getAddress())
												+ ": " + additional), 1, TimeUnit.MINUTES);	

				if (debug)
					System.out
							.println("\n--------------------------------------DEBUG INFO START (doCommand, send)  --------------------------------------"
									+ "\nNachricht '" + additional + "' wurde versandt.\n"
									+ "--------------------------------------DEBUG INFO END (doCommand, send)  --------------------------------------\n");
			}
		}
	}
	
	// Hilfemethode für das Senden einer Nachricht: Liefert die Root-Wavelet einer Wave zurück
	public WaveletData getOpenWavelet() {
		return (openWave == null) ? null : ClientUtils.getConversationRoot(this.openWave);
	}
	
	// Hilfemethode für das Senden einer Nachricht
	private BufferedDocOp getManifestDocument() {
		return getOpenWavelet() == null ? null : getOpenWavelet().getDocuments().get(DocumentConstants.MANIFEST_DOCUMENT_ID);
	}

	/**
	 * Diese Methode wird immer dann aufgerufen, wenn Änderungen an einer Wave vorgenommen werden.
	 * 
	 * @param arg0	Wavelet Informationen
	 */
	@Override
	public void onDeltaSequenceEnd(WaveletData arg0) {
		if (debug) {
			System.out.println("\n-------------------------------------- DEBUG INFO START (onDeltaSequenceEnd)  --------------------------------------"
					+ "\n"
					+ this.debugCounter
					+ ". Aufruf der Methode onDeltaSequenceEnd");
			System.out.println("Wavelet Daten: " + arg0);
			System.out.println("--------------------------------------  DEBUG INFO END (onDeltaSequenceEnd)  --------------------------------------\n");
			this.debugCounter++;
		}
	}

	/**
	 * Diese Methode wird immer dann aufgerufen, wenn jemand einer Wave hinzugefügt wird.
	 * 
	 * @param arg0 Wer möchte hinzufügen?
	 * @param arg1 Wer soll hinzugefügt werden?
	 * @param arg2 In welcher Wave wird hinzugefügt?
	 */
	@Override
	public void participantAdded(String arg0, WaveletData arg1, ParticipantId arg2) {
		if (debug) {
			System.out.println("\n-------------------------------------- DEBUG INFO START (participantAdded)  --------------------------------------"
					+ "\n"
					+ this.debugCounter
					+ ". Aufruf der Methode participantAdded");
			System.out.println("Wer fügt hinzu? " + arg0);
			System.out.println("Wer soll hinzugefügt werden? " + arg2.toString());
			System.out.println("In welcher Wave wird hinzugefügt? " + arg1.getWaveletName().waveId.getId());
			System.out.println("--------------------------------------  DEBUG INFO END (participantAdded)  --------------------------------------\n");
			this.debugCounter++;
		}
		
		// Nur wenn jemand anderes einer Wave hinzugefügt wird, soll dessen Waveliste aktualisiert werden
		if(!arg0.equals(arg2) && arg1.getWaveletName().waveId.getId().startsWith("w+")) {
			clientGUI.displaySignInGUI.getDefault().asyncExec(
					new Runnable() {
						public void run(){
							getMyWaves();
						}
					});
		}
	}

	/**
	 * Diese Methode wird immer dann aufgerufen, wenn jemand in einer Wave gelöscht wird.
	 * 
	 * @param arg0 Wer möchte löschen?
	 * @param arg1 Wer soll gelöscht werden?
	 * @param arg2 Aus welcher Wave wird gelöscht?
	 */
	@Override
	public void participantRemoved(String arg0, WaveletData arg1, ParticipantId arg2) {
		if (debug) {
			System.out.println("\n-------------------------------------- DEBUG INFO START (participantRemoved)  --------------------------------------"
					+ "\n"
					+ this.debugCounter
					+ ". Aufruf der Methode participantRemoved");
			System.out.println("Wer möchte löschen? " + arg0);
			System.out.println("Wer soll gelöscht werden? " + arg2.getAddress().toString());
			System.out.println("Aus welcher Wave wird gelöscht? " + arg1.getWaveletName().waveId.getId());
			System.out.println("--------------------------------------  DEBUG INFO END (participantRemoved)  --------------------------------------\n");
			this.debugCounter++;
		}
		
		// Nur wenn jemand anderes einer Wave entzogen wird, soll dessen Waveliste aktualisiert werden
		if(!arg0.equals(arg2) && arg1.getWaveletName().waveId.getId().startsWith("w+")) {
			clientGUI.displaySignInGUI.getDefault().asyncExec(
					new Runnable() {
						public void run(){
							// Auch an dieser Stelle muss leider 50 Millisekunden gewartet werden, damit das Backend aktuelle Informationen liefert
							try {
								Thread.sleep(50);
							} catch (InterruptedException e1) {
								e1.printStackTrace();
							}
							
							getMyWaves();
						}
					});
		}
	}

	/**
	 * Diese Methode wird immer dann aufgerufen, wenn ein Wave Dokument erstellt bzw. geändert wird.
	 * 
	 * @param arg0 Wer führt Änderungen durch?
	 * @param arg1 WaveletData
	 * @param arg2 WaveletDocumentOperation
	 */
	@Override
	public void waveletDocumentUpdated(String arg0, WaveletData arg1, WaveletDocumentOperation arg2) {
		if (debug) {
			System.out.println("\n-------------------------------------- DEBUG INFO START (waveletDocumentUpdated)  --------------------------------------"
					+ "\n"
					+ this.debugCounter
					+ ". Aufruf der Methode waveletDocumentUpdated");
			System.out.println("Wer führt Änderungen durch? " + arg0);
			System.out.println("WaveletData: " + arg1);
			System.out.println("WaveletDocumentOperation: " + arg2);
			System.out.println("--------------------------------------  DEBUG INFO END (waveletDocumentUpdated)  --------------------------------------\n");
			this.debugCounter++;
		}
		
		// Die zu aktualisierende Wave ist geöffnet
		if(this.openWave != null) {
			// Die WaveID der geöffneten Wave spricht mit der zu aktualisierenden WaveID überein
			if(this.openWave.getWaveId().getId().equals(arg1.getWaveletName().waveId.getId())) {
				if(!arg2.getOperation().getCharactersString(3).isEmpty()) {
					// Dem Nachrichtenfeld wird die neueste Nachricht angehängt
					clientGUI.displaySignInGUI.getDefault().asyncExec(
							new MyRunnable(arg2.getOperation().getCharactersString(3)) {
								public void run(){
									clientGUI.textWaveContent.append(super.filterOrUpdateText + "\n");
								}
							});
				}
			}
		}
	}
	
	/**
	 * Diese Methode initialisiert mir meine GUI mit den Startwerten. D.h. Selektierung aller vorhandenen Waves und meiner Kontakte.
	 */
	public void initializeGUI(){
		myRunnable = new MyRunnable("initializeGUI");
		myRunnable.run();
	}

	/**
	 * Diese Methode gibt Auskunft darüber, ob mein Client noch mit dem Backend verbunden ist.
	 * 
	 * @return boolean
	 */
	private boolean clientIsConnected(){
		if(this.backend != null)
			return true;
		else
			return false;
	}
	
	/**
	 * Diese Methode entfernt den Domänennamen für die Darstellung der Teilnehmer. Beachte: Für das Backend sind die Domänennamen mit anzugeben.
	 * @param adress
	 * @return
	 */
	public String transformAdress(String address){
		return address.substring(0, address.indexOf("@")).toLowerCase();
	}
	
	/**
	 * Diese Methode selektiert alle mir zugeordneten Waves aus dem Backend.
	 */
	private void getMyWaves() {
		clientGUI.listWaveList.removeAll();
		myWaves = ClientUtils.getIndexEntries(backend.getIndexWave());
		
		// Nur wenn Waves vorhanden sind, werden diese ausgegeben
		if(myWaves.size() > 0) {
			for (IndexEntry myWave : myWaves) {
				clientGUI.listWaveList.add("Wave: " + myWave.getWaveId().getId());
			}
		} else {
			clientGUI.listWaveList.add("Es liegen keine Waves vor.");
		}
	}
	
	/**
	 * Diese Methode dient dem Hinzufügen eines Benutzers zu einer bestimmten Wave.
	 */
	public void addParticipant(String name) {
		ParticipantId newParticipant = new ParticipantId(name);

		if (!getOpenWavelet().getParticipants().contains(newParticipant)) {
			backend.sendAndAwaitWaveletOperation(getOpenWavelet().getWaveletName(), new AddParticipant(newParticipant), 1, TimeUnit.MINUTES);
			
			if (debug)
				System.out
						.println("\n--------------------------------------  DEBUG INFO START (addParticipant)  --------------------------------------"
								+ "\nBenutzer " + name + " wurde der Wave erfolgreich hinzugefügt.\n"
								+ "--------------------------------------  DEBUG INFO END (addParticipant)  --------------------------------------\n");
		}
	}
	
	/**
	 * Diese Methode dient dem Entfernen eines Benutzers von einer bestimmten Wave.
	 */
	public void removeParticipant(String name) {
		ParticipantId newParticipant = new ParticipantId(name);

		if (getOpenWavelet().getParticipants().contains(newParticipant)) {
			backend.sendAndAwaitWaveletOperation(getOpenWavelet().getWaveletName(), new RemoveParticipant(newParticipant), 1, TimeUnit.MINUTES);

			// Wenn ich selbst aus der zurzeit geöffneten Wave entfernt werde, muss das openWave Objekt auch auf null gesetzt werden
			if(backend.getUserId().getAddress().equals(name)) this.openWave = null;
			
			if (debug)
				System.out
						.println("\n--------------------------------------  DEBUG INFO START (removeParticipant)  --------------------------------------"
								+ "\nBenutzer " + name + " wurde von der Wave erfolgreich gelöscht.\n"
								+ "--------------------------------------  DEBUG INFO END (removeParticipant)  --------------------------------------\n");
		}
	}
	
	/**
	 * Vom Framework zurzeit keine Verwendung.
	 */
	@Override
	public void onDeltaSequenceStart(WaveletData arg0) {}
	
	/**
	 * Vom Framework zurzeit keine Verwendung.
	 */
	@Override
	public void noOp(String arg0, WaveletData arg1) {}

	/**
	 * Vom Framework zurzeit keine Verwendung.
	 */
	@Override
	public void onCommitNotice(WaveletData arg0, HashedVersion arg1) {}
}