import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * 
 * Diese Klasse repräsentiert die eigentliche Feasibility Prototype GUI.
 * 
 * @author Jan Schwarzer
 *
 */
public class ClientGUI {
	
	// Der Wave Client
	public Client waveClient = null;

	// Der GUI-Rahmen
	public Shell shellGUI = null;
	
	// Das GUI-Display
	public Display displaySignInGUI = null;

	// Die GUI-Abmessungen
	public int ShellGUIHeight = 480;
	public int ShellGUIWidth = 700;

	// Labels für den Login
	public Label labelWaveList = null;
	public Label labelWaveContent = null;
	public Label labelContacts = null;
	public Label labelParticipants = null;

	// Textfelder
	public Text textWaveContent = null;
	public Text textPost = null;

	// Listen zur Selektierung
	public List listWaveList = null;
	public List listContacts = null;
	public List listParticipants = null;

	// Buttons
	public Button buttonNewWave = null;
	public Button buttonPost = null;
	public Button buttonOpenWave = null;

	public MouseListener mouseListenerParticipants = null;

	public MouseListener createParticipantMouseListener() {
		mouseListenerParticipants = new MouseListener() {

			@Override
			public void mouseUp(MouseEvent arg0) {}

			@Override
			public void mouseDown(MouseEvent arg0) {}

			@Override
			public void mouseDoubleClick(MouseEvent arg0) {
				
				// Auswahl von Teilnehmern, nur wenn auch eine Wave geöffnet ist
				if(waveClient.openWave != null) {
					// Jemand möchte einen Kontakt hinzufügen (reagiert nur auf Doppelklick)
					if(listContacts.isFocusControl()) {
						if(listContacts.getItemCount() > 0 && listContacts.getSelectionIndex() >= 0) {
							boolean isAlreadyParticipant = false;

							for(String participant : listParticipants.getItems()){
								if(participant.equals(listContacts.getItem(listContacts.getSelectionIndex())))
									isAlreadyParticipant = true;
							}
							
							// Nur wenn der Benutzer noch nicht der Liste der Teilnehmer einer Wave angehört, kann dieser hinzugefügt werden
							if(!isAlreadyParticipant) {
								listParticipants.add(listContacts.getItem(listContacts.getSelectionIndex()));
								waveClient.addParticipant(listContacts.getItem(listContacts.getSelectionIndex()) + domainName);
							}
						}
						
					// Jemand möchte einen Kontakt löschen
					} else if(listParticipants.isFocusControl()) {
						if(listParticipants.getItemCount() > 0 && listParticipants.getSelectionIndex() >= 0) {
							waveClient.removeParticipant(listParticipants.getItem(listParticipants.getSelectionIndex()) + domainName);
							listParticipants.remove(listParticipants.getSelectionIndex());
						}
						
					// Alle anderen Fälle interssieren in diesem speziellen Fall nicht (reagiert nur auf Doppelklick)
					} else {}
				}
			}
		};

		return mouseListenerParticipants;
	}
	
	// Wird von SingInGUI übergeben
	public String domainName = null;

	public ClientGUI(Client wc, Display displaySignInGUI, String domainName) {
		waveClient = wc;
		waveClient.clientGUI = this;
		this.domainName = domainName;

		shellGUI = new Shell(displaySignInGUI);
		displaySignInGUI = displaySignInGUI;

		shellGUI.setText("Willkommen " + waveClient.transformAdress(waveClient.backend.getUserId().getAddress()));
		shellGUI.setSize(ShellGUIWidth, ShellGUIHeight);
		shellGUI.addShellListener(new ShellAdapter() {
			
			public void shellClosed(ShellEvent e) {
				shellGUI.dispose();
			}

		});
		
		labelWaveList = new Label(shellGUI, SWT.NONE);
		labelWaveList.setText("Meine Waves:");
		labelWaveList.setSize(150, 20);
		labelWaveList.setLocation(10, 10);

		labelWaveContent = new Label(shellGUI, SWT.NONE);
		labelWaveContent.setText("Nachrichten:");
		labelWaveContent.setSize(150, 20);
		labelWaveContent.setLocation(350, 10);

		labelContacts = new Label(shellGUI, SWT.NONE);
		labelContacts.setText("Meine Kontakte:");
		labelContacts.setSize(150, 20);
		labelContacts.setLocation(10, 350);

		labelParticipants = new Label(shellGUI, SWT.NONE);
		labelParticipants.setText("Teilnehmer der Wave:");
		labelParticipants.setSize(150, 20);
		labelParticipants.setLocation(350, 350);

		textWaveContent = new Text(shellGUI, SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
		// startX, startY, Breite, Höhe
		textWaveContent.setBounds(350, 35, 340, 200);
		textWaveContent.setEditable(false);

		textPost = new Text(shellGUI, SWT.BORDER | SWT.V_SCROLL);
		textPost.setBounds(350, 250, 340, 50);

		listWaveList = new List(shellGUI, SWT.MULTI | SWT.H_SCROLL);
		// startX, startY, Breite, Höhe
		listWaveList.setBounds(10, 35, 330, 200);

		listContacts = new List(shellGUI, SWT.MULTI | SWT.H_SCROLL);
		listContacts.setBounds(10, 375, 200, 70);
		listContacts.addMouseListener(this.createParticipantMouseListener());

		listParticipants = new List(shellGUI, SWT.MULTI | SWT.H_SCROLL);
		listParticipants.setBounds(350, 375, 200, 70);
		listParticipants.addMouseListener(this.createParticipantMouseListener());

		buttonNewWave = new Button(shellGUI, SWT.PUSH);
		buttonNewWave.setText("Neue Wave");
		buttonNewWave.setLocation(10, 250);
		buttonNewWave.setSize(100, 30);

		buttonPost = new Button(shellGUI, SWT.PUSH);
		buttonPost.setText("Nachricht senden");
		buttonPost.setLocation(350, 300);
		buttonPost.setSize(150, 30);

		buttonOpenWave = new Button(shellGUI, SWT.PUSH);
		buttonOpenWave.setText("Wave öffnen");
		buttonOpenWave.setLocation(110, 250);
		buttonOpenWave.setSize(100, 30);

		// Erstellen einer neuen Wave
		buttonNewWave.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				waveClient.doCommand("new", "");
			}
		});

		// Senden einer Nachricht
		buttonPost.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (!textPost.getText().isEmpty())
					waveClient.doCommand("send", textPost.getText());

				textPost.setText("");
			}
		});

		// Öffnen einer Wave
		buttonOpenWave.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (listWaveList.getFocusIndex() > -1) {
					waveClient.doCommand("open", "");
				}
			}
		});

		// GUI öffnen
		shellGUI.open();
		
		// Initializierung der GUI mit Werten aus dem Backend
		waveClient.initializeGUI();
		
		while (!shellGUI.isDisposed()) {
			if (!displaySignInGUI.readAndDispatch())
				displaySignInGUI.sleep();
		}
	}

}
